;;;; -*- Mode:Lisp; Syntax: Common-Lisp; Package:UMASS-EXTENDED-LISP; Base:10 -*-
;;;; *-* File: Dawg:DIS$DISK:[GBB.V-120.LOCAL]DEC-EXTENSIONS.LISP *-* 
;;;; *-* Last-Edit: Thursday, July 12, 1990  13:48:18 *-*
;;;; *-* Edited-By: Gallagher *-*
;;;; *-* Machine: Gilgamesh (Explorer II, Microcode 429) *-*
;;;; *-* Software: TI Common Lisp System 6.31 *-*
;;;; *-* Lisp: TI Common Lisp System 6.31 *-*

;;;; **************************************************************************
;;;; **************************************************************************
;;;; *
;;;; *            VAX Lisp Implementation Dependent Extensions
;;;; *
;;;; **************************************************************************
;;;; **************************************************************************
;;;
;;; Written by: Kevin Gallagher/Philip Johnson/Daniel Corkill/Kelly Murray
;;;             Department of Computer and Information Science
;;;             University of Massachusetts
;;;             Amherst, Massachusetts 01003.
;;;
;;; This code was written as part of the GBB (Generic Blackboard) system at
;;; the Department of Computer and Information Science (COINS), University of 
;;; Massachusetts, Amherst.
;;;
;;; Copyright (c) 1986-1990 COINS.  
;;; All rights reserved.
;;;
;;; Development of this code was partially supported by:
;;;    NSF CER grant DCR-8500332;
;;;    Donations from Texas Instruments, Inc.;
;;;    ONR URI grant N00014-86-K-0764.
;;;
;;; Permission to copy this software, to redistribute it, and to use it for
;;; any purpose is granted, subject to the following restrictions and
;;; understandings.
;;;
;;; 1.  Title and copyright to this software and any material associated
;;; therewith shall at all times remain with COINS.  Any copy made of this
;;; software must include this copyright notice in full.
;;;
;;; 2.  The user acknowledges that the software and associated materials
;;; are provided as a research tool that remains under active development
;;; and is being supplied ``as is'' for the purposes of scientific
;;; collaboration aimed at further development and application of the
;;; software and the exchange of technical data.
;;;
;;; 3.  All software and materials developed as a consequence of the use of
;;; this software shall duly acknowledge such use, in accordance with the
;;; usual standards of acknowledging credit in academic research.
;;;
;;; 4.  Users of this software agree to make their best efforts to inform
;;; the COINS GBB Development Group of noteworthy uses of this software.
;;; The COINS GBB Development Group can be reached at:
;;;
;;;     GBB Development Group
;;;     C/O Dr. Daniel D. Corkill
;;;     Department of Computer and Information Science
;;;     Lederle Graduate Research Center
;;;     University of Massachusetts
;;;     Amherst, Massachusetts 01003
;;;
;;;     (413) 545-0156
;;;
;;; or via electronic mail:
;;;
;;;     GBB@CS.UMass.Edu
;;;
;;; Users are further encouraged to make themselves known to this group so
;;; that new releases, bug fixes, and tutorial information can be
;;; distributed as they become available.
;;;
;;; 5.  COINS makes no representations or warranties of the merchantability
;;; or fitness of this software for any particular purpose; that uses of
;;; the software and associated materials will not infringe any patents,
;;; copyrights, trademarks, or other rights; nor that the operation of this
;;; software will be error-free.  COINS is under no obligation to provide
;;; any services, by way of maintenance, update, or otherwise.  
;;;
;;; 6.  In conjunction with products or services arising from the use of
;;; this material, there shall be no use of the name of the Department of
;;; Computer and Information Science or the University of Massachusetts in
;;; any advertising, promotional, or sales literature without prior written
;;; consent from COINS in each case.
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;;;
;;;  This file contains extensions to Common Lisp needed by GBB.  The functions
;;;  in this file depend on the details of the VAXLisp implementation.  The
;;;  file UMASS-EXTENDED-LISP.LISP contains extensions to Common Lisp that are
;;;  written in Common Lisp.
;;;
;;;  03-25-87 File Created.  (Gallagher)
;;;  03-01-90 Updated to VaxLisp 3.0.  (Gallagher)
;;;
;;; * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

(in-package 'umass-extended-lisp
	    :use '(lisp))

(export '(defstruct-p
	  defstruct-conc-name
	  structure-slot-names
	  structure-slot-p
	  get-structure-slot
	  delete-transform
	  %pointer))

(export '(deftransform
	  %pass%))


(use-package '(lisp))

(proclaim '(optimize (speed 3) (safety 1)))

;;; The functions that must be defined in this file are:
;;;
;;;     defstruct-p
;;;     get-defstruct-description
;;;     defstruct-conc-name
;;;     structure-slot-names
;;;     structure-slot-p
;;;     get-structure-slot  (and corresponding setf method)
;;;     deftransform
;;;     %pointer
;;;
;;; See the documentation with each function for arguments and purpose.


(eval-when (compile load eval)
  (cond ((string-equal "V3" (lisp-implementation-version) :end2 2)
         (pushnew :VAXLISP-V3 *features*))
        (t
         (warn "DEC-EXTENSIONS hasn't been tested in VaxLisp ~a.~@
                 Try it out to see if it works and update this file."
                (lisp-implementation-version)))))

;; Slot descriptor access funtions:

(defun sd-slot-name (sd)
  (svref sd 0))

(defun sd-slot-offset (sd)
  (svref sd 1))



;;;; --------------------------------------------------------------------------
;;;;   Generic Defstruct Functions
;;;; --------------------------------------------------------------------------

;;; It is a hole in Common Lisp that you can't access a slot in a structure
;;; via the slot name.  For example, if I have an instance of the following
;;; defstruct,
;;;
;;;    (defstruct (foo (:conc-name FOO->))
;;;      (a nil)
;;;      (b nil))
;;;
;;; I want to be able to get the value of slot A with a call like:
;;;
;;;    (get-structure-slot object 'A).
;;;
;;; This is because I may not know what the accessor function is for that
;;; slot.  Even if I did know the name of the accessor function, I wouldn't
;;; be able to use it to alter the value of the slot because Setf needs to
;;; know the function name at compile time.  This means that code such as
;;; the following won't work.
;;;
;;;    (defun set-structure-slot (object slot-accessor new-value)
;;;      (setf (funcall slot-accessor object) new-value))
;;;
;;; Obviously, setf shouldn't need to handle that as a place form.  This
;;; just points out the need for a general slot accessor function which
;;; will work as a place form for setf.

;;; GBB uses the functions Get-Structure-Slot and Defstruct-Conc-Name
;;; which are defined below.


(defun get-defstruct-description (name)

  "GET-DESFSTRUCT-DESCRIPTION name

   If `Name' has been defined as a structure then return its
   description.  Otherwise signal an error."

  (if (subtypep name 'system::structure)
      (system::find-type name)
      (error "~s is not the name of a defstruct." name)))

(defun defstruct-p (name)

  "DEFSTRUCT-P name

   This predicate returns a true if `Name' is the name of a
   defstruct type; nil otherwise.  `Name' must be a symbol."

  (subtypep name 'system::structure))


(defun defstruct-conc-name (structure)

  "DEFSTRUCT-CONC-NAME structure

   Return a string which is the prefix for building defstruct slot
   accessors for this structure.  STRUCTURE must be a symbol which
   names a structure type."

  (system::type-conc-name (system::find-type structure)))

(defun structure-slot-names (type)

  "STRUCTURE-SLOT-NAMES type

   This function returns a list of structure slot names
   given the type (symbol) of the structure."

  (let* ((description (get-defstruct-description type))
	 (slot-data (system::type-slot-descriptors description)))
    (map 'list #'sd-slot-name (the simple-vector slot-data))))


(defun structure-slot-p (type slot)

  "STRUCTURE-SLOT-P type slot

   This function returns true if `Slot' is the name of a slot in the
   defstruct type named by `Type' and nil if not."

  (and (structure-slot-index-1 type slot)
       t))


;;; Structure Slot Index Computation Function:
;;;
;;; This is common to both the access and the setting functions.  It
;;; computes the index into the internal data structure (usually a
;;; vector) where the slot value is stored.


(defun structure-slot-index (type slot)

  ;; Returns the index of the slot or signals an error
  ;; if the slot doesn't exist.
  
  (or (structure-slot-index-1 type slot)
      (error "~S is not a slot in ~S." slot type)))


(defun structure-slot-index-1 (type slot)

  ;; Returns the index of the slot or nil if the slot doesn't
  ;; exist.  Used by STRUCTURE-SLOT-P and STRUCTURE-SLOT-INDEX.

  (let* ((description (get-defstruct-description type))
         (slot-data (system::type-slot-descriptors description))
         (sd (find slot slot-data :key #'sd-slot-name :test #'string=)))
    (and sd (sd-slot-offset sd))))
    

;;;
;;; Structure Slot Access Functions :::
;;;

(defun get-structure-slot (object slot)

  "GET-STRUCTURE-SLOT object slot

Returns the value of a slot in a structure
given the name of the slot.

This function may be used as a place form for SETF."

  (let ((type (type-of object)))
    (system::%sp-struct-ref
      type object (structure-slot-index type slot))))

;;;
;;; Structure Slot Setting Functions :::
;;;

(defun set-structure-slot (object slot value)

  ;; SET-STRUCTURE-SLOT object slot value
  ;;
  ;; Sets the value of a slot in a structure
  ;; given the name of the slot."

  (let ((type (type-of object)))
    (system::%sp-struct-set
      type object (structure-slot-index type slot) value)))

(defsetf get-structure-slot set-structure-slot)


;;;; --------------------------------------------------------------------------
;;;;   Compiler Transforms.
;;;; --------------------------------------------------------------------------

;;; The body of the transform should return the symbol %PASS% if it
;;; doesn't apply.  If it does apply then it should return the
;;; transformed form.  For example, a common optimization is to do
;;; keyword analysis at compile time.
;;;
;;;     (defun foo (x &key a b)
;;;       ...)
;;;
;;;     (deftransform (foo keyword) (x &rest keyword-args)
;;;       (if (literal-keywords-in-arglist keyword-args)
;;;           (let ((a (getf keyword-args :a))
;;;                 (b (getf keyword-args :b)))
;;;             `(foo-internal ,x ,a ,b))
;;;           '%pass%))
;;;
;;;     (defun foo-internal (x a b)
;;;       ...)

;; Not implemented yet in Vaxlisp 3.1

(defmacro deftransform ((fn name) args &body body)
  fn name args body
  `(progn))

(defun delete-transform (function transform-name)

  "DELETE-TRANSFORM function transform-name

   This function will remove the compiler transform for FUNCTION
   designated by TRANSFORM-NAME."

  nil)


;;;; --------------------------------------------------------------------------
;;;;   Miscellaneous Non Common Lisp Functions
;;;; --------------------------------------------------------------------------


;;; This function is used when printing internal GBB objects.  If your
;;; implementation doesn't provide an easy way to get the address of
;;; an object it is ok to simply return zero.  GBB doesn't depend on the
;;; value of this function uniquely identifying an object.

(defun %pointer (x)

  "%POINTER x

   Returns the address of X.  This is only used in printing to easily 
   distiguish two objects.  Its value should not be relied on."

  (system::address-of x))


;;; ---------------------------------------------------------------------------
;;;				  End of File
;;; ---------------------------------------------------------------------------
